#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2023-01-17 21:55:23
# @Author  : HeLiang (helianghit@foxmail.com)
# @Link    : https://github.com/HeLiangHIT

import os
from ui.qt_main import main

if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2023-01-17 21:55:38
# @Author  : HeLiang (helianghit@foxmail.com)
# @Link    : https://github.com/HeLiangHIT

'''
迷宫算法涉及的基本数据结构，包括Point和Map，使用示例:
>>> m = Map(); print(m) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
10X10:
C C C C C C C C C C
    ...
>>> m.data[1][1] = Point.Wall
>>> m.data = Point.Wall # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
Traceback (most recent call last):
    ...
AttributeError: can't set attribute 'data'
>>> print(m) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
10X10:
C C C C C C C C C C
C W C C C C C C C C
    ...
>>> m.save("./empty.map")
>>> m1 = Map.load("./empty.map")
>>> print(m1) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
10X10:
C C C C C C C C C C
C W C C C C C C C C
    ...
>>> os.remove("./empty.map")
>>> m1.data[0][0] = Point.Start;m1.data[9][9] = Point.End;print(Map().diff(m1))
[(0, 0, <Point.Start: 3>), (1, 1, <Point.Wall: 2>), (9, 9, <Point.End: 4>)]
>>> print(m1.start, m1.end)
[(0, 0)] [(9, 9)]
>>> txt = m1.to_json(); print(txt)
{"data": [[3, 1, 1, 1, 1, 1, 1, 1, 1, 1], [1, 2, 1, 1, 1, 1, 1, 1, 1, 1], [1, 1, 1, 1, 1, 1, 1, 1, 1, 1], [1, 1, 1, 1, 1, 1, 1, 1, 1, 1], [1, 1, 1, 1, 1, 1, 1, 1, 1, 1], [1, 1, 1, 1, 1, 1, 1, 1, 1, 1], [1, 1, 1, 1, 1, 1, 1, 1, 1, 1], [1, 1, 1, 1, 1, 1, 1, 1, 1, 1], [1, 1, 1, 1, 1, 1, 1, 1, 1, 1], [1, 1, 1, 1, 1, 1, 1, 1, 1, 4]]}
>>> m2, isok = Map().from_json(txt); print(m2) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
10X10:
S C C C C C C C C C
C W C C C C C C C C
...
'''
import os
import pickle
import json
import copy
from enum import IntEnum


Point = IntEnum('Point', ('Chan', 'Wall', 'Start', 'End', "Visited", "NxtVisit")) # Enum is not JSON serializable


class Map(object):
    '''地图对象: 支持创建、修改成员、保存、加载'''
    def __init__(self, size = (10, 10), default = Point.Chan, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._size = size
        self._data = [[default for _ in range(self._size[1])] for _ in range(self._size[0])]

    def __str__(self):
        # mtx = '\n'.join([' '.join([f"{p.value}" for p in row]) for row in self._data])
        mtx = '\n'.join([' '.join([p.name[0] for p in row]) for row in self._data])
        return f"{self._size[1]}X{self._size[0]}:\n{mtx}"

    def from_data(self, data):
        self._size = (len(data), len(data[0]))
        self._data = copy.deepcopy(data)
        return self

    def diff(self, other): # different from __sub__ witch return Map
        # 计算差异: [(x, y, other_value), ...]
        res = []
        if self._size != other.size:
            print(f"Map diff with different size: {self.size} != {other.size}")
            return res
        for m in range(self._size[0]):
            for n in range(self._size[1]):
                if self._data[m][n] != other.data[m][n]:
                    res.append((m, n, other.data[m][n]))
        return res

    @property # avoid modify data use m.data = Point.Wall, can use as m.data[1][1] = Point.Wall
    def data(self):
        return self._data

    @property
    def size(self):
        return self._size

    @property
    def start(self):
        res = []
        for m in range(self._size[0]):
            for n in range(self._size[1]):
                if self._data[m][n] == Point.Start:
                    res.append((m, n))
        return res

    @property
    def end(self):
        res = []
        for m in range(self._size[0]):
            for n in range(self._size[1]):
                if self._data[m][n] == Point.End:
                    res.append((m, n))
        return res

    def save(self, path):
        with open(path, "wb") as f:
            s = pickle.dumps(self)
            f.write(s)

    @staticmethod
    def load(path):
        with open(path, "rb") as f:
            s = f.read()
            m = pickle.loads(s)
            return m

    def to_json(self):
        txt = json.dumps({
            "data": self._data
            })
        return txt

    @staticmethod
    def from_json(txt):
        dic = json.loads(txt)
        values = tuple(item.value for item in Point) # 用于校验
        if not all([p in values for row in dic["data"] for p in row]): # 确保所有值都是有效Point
            print(f"json value is not valid Point: {dic['data']}")
            return None, False
        data = [[Point(p) for p in row] for row in dic["data"]]
        return Map().from_data(data), True


if __name__ == '__main__':
    import doctest
    doctest.testmod()  # verbose=True shows the output


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2023-01-18 16:50:55
# @Author  : HeLiang (helianghit@foxmail.com)
# @Link    : https://github.com/HeLiangHIT
# @Ref     : https://blog.csdn.net/juzihongle1/article/details/73135920 和 https://www.cnblogs.com/shiroe/p/15506909.html
#            还有递归分割等很多迷宮生成算法
# 此处r主要是为了支持在 doctest 命令中使用 \n
r'''
迷宫生成，提供dfsg(深度优先)和primg(随机广度)两种方法，入参相同，使用示例如下:
>>> callback = lambda m: print('\n'.join(["-----", ] + [' '.join([p.name[0] for p in row]) for row in m]))
>>> m = dfsg((3, 3), [(0, 0),], [(2, 2),], callback=callback) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
-----
C W W
...
...C
>>> print(m) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
3X3:
S...
...
...E
>>> m = primg((3, 3), [(0, 0),], [(2, 2),], callback=callback) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
-----
C W W
...
...C
>>> print(m) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
3X3:
S...
...
...E
>>> m = dfsg((6, 6), [(0, 0),], [(5, 5),], callback=callback) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
-----
C...
...
...W
>>> print(m) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
6X6:
S...
...
...E
>>> m = primg((60, 60), [(0, 0),], [(45, 51),]) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
>>> print(m) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
60X60:
S...
...
W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W W
'''
import sys
import copy
import random
import traceback
from functools import reduce
from .map import Point, Map


def _valid_check(size, starts, ends):
    '''输入有效性检查'''
    if not (isinstance(starts, list) and isinstance(starts[0], tuple) and isinstance(starts[0][0], int)):
        print(f"input parameter starts is illegal(must be list of coordinate): {starts}")
        traceback.print_stack()
        return False
    if not (isinstance(ends, list) and isinstance(ends[0], tuple) and isinstance(ends[0][0], int)):
        print(f"input parameter ends is illegal(must be list of coordinate): {ends}")
        traceback.print_stack()
        return False
    # TODO: srcs/targets 是否超过边界
    return True


def _valid_neighbour(m, cur, value = Point.Wall):
    size = m.size
    udlr = [(cur[0] - 1, cur[1]), (cur[0] + 1, cur[1]), (cur[0], cur[1] - 1), (cur[0], cur[1] + 1),]
    unvisited_nb = reduce(lambda x, y: x + y, [0 <= q[0] < size[0] and 0 <= q[1] < size[1] and m.data[q[0]][q[1]] == value for q in udlr])
    return unvisited_nb


def _connect_points(m, targets):
    for p in targets:
        if _valid_neighbour(m, p, Point.Chan) == 0:
            udlr = [(p[0] - 1, p[1]), (p[0] + 1, p[1]), (p[0], p[1] - 1), (p[0], p[1] + 1),]
            ava_nb = filter(lambda x: _valid_neighbour(m, x, Point.Chan) >= 1, udlr)
            link = random.choice(list(ava_nb))
            m.data[link[0]][link[1]] = Point.Chan
    return m


def dfsg(size, starts, ends, callback = None):
    '''深度优先迷宫生成算法:
    算法流程:
    1.将起点作为当前迷宫单元并标记为已访问
    2.当还存在未标记的迷宫单元，进行循环
        1.如果当前迷宫单元有未被访问过的相邻的迷宫单元
            1.随机选择一个未访问的相邻迷宫单元
            2.将当前迷宫单元入栈
            3.移除当前迷宫单元与相邻迷宫单元的墙
            4.标记相邻迷宫单元并用它作为当前迷宫单元
        2.如果当前迷宫单元不存在未访问的相邻迷宫单元，并且栈不空
            1.栈顶的迷宫单元出栈
            2.令其成为当前迷宫单元
    这种算法生成的迷宫会有比较明显的主路
    '''
    if not _valid_check(size, starts, ends):
        return []
    pre_stack_limit = sys.getrecursionlimit()
    sys.setrecursionlimit(10000)
    m = Map(size, default = Point.Wall)
    visited_flag = [[False for _ in range(size[1])] for _ in range(size[0])]

    def dfs(cur):
        visited_flag[cur[0]][cur[1]] = True
        m.data[cur[0]][cur[1]] = Point.Chan

        if callback is not None:
            callback(m.data)

        if cur in ends: # 走到终点就不用打通终点周围的墙了
            return
        udlr = [(cur[0] - 2, cur[1]), (cur[0] + 2, cur[1]), (cur[0], cur[1] - 2), (cur[0], cur[1] + 2),]
        random.shuffle(udlr)
        for p in udlr:
            if 0 <= p[0] < size[0] and 0 <= p[1] < size[1] and not visited_flag[p[0]][p[1]] and _valid_neighbour(m, p) >= 1:
                if cur[0] == p[0]:
                    m.data[p[0]][min(cur[1], p[1]) + 1] = Point.Chan
                    visited_flag[p[0]][min(cur[1], p[1]) + 1] = True
                else:
                    m.data[min(cur[0], p[0]) + 1][p[1]] = Point.Chan
                    visited_flag[min(cur[0], p[0]) + 1][p[1]] = True
                dfs(p)

    dfs(starts[0]) # DFS 只需要一条线走到黑
    # start[0] 和 starts[1:]/end 之间差非偶数的情况下无法联通，需要链接终点到最近的可行点
    m = _connect_points(m, starts[1:] + ends)

    for p in starts:
        m.data[p[0]][p[1]] = Point.Start
    for p in ends:
        m.data[p[0]][p[1]] = Point.End
    sys.setrecursionlimit(pre_stack_limit)
    return m


def primg(size, starts, ends, callback = None):
    '''随机Prim算法生成迷宫 - 更像随机广度优先算法
    1.让迷宫全是墙
    2.选一个单元格作为迷宫的通路，然后把它的邻墙放入列表
    3.当列表里还有墙时
        1.从列表里随机选一个墙，如果这面墙分隔的两个单元格只有一个单元格被访问过
            1.那就从列表里移除这面墙，即把墙打通，让未访问的单元格成为迷宫的通路
            2.把这个格子的墙加入列表
        2.如果墙两面的单元格都已经被访问过，那就从列表里移除这面墙
    改进: 如果把墙放到列表中，比较复杂，维基里面提到了改进策略。可以维护一个迷宫单元格的列表，而不是边的列表。在这个迷宫单元格列表里面存放了未访问的单元格，我们在单元格列表中随机挑选一个单元格，如果这个单元格有多面墙联系着已存在的迷宫通路，我们就随机选择一面墙打通。
    相对于深度优先的算法，Prim随机算法不是优先选择最近选中的单元格，而是随机的从所有的列表中的单元格进行选择，新加入的单元格和旧加入的单元格同样概率会被选择，新加入的单元格没有有优先权。因此其分支更多，生成的迷宫更复杂，难度更大，也更自然。
    '''
    if not _valid_check(size, starts, ends):
        return []
    m = Map(size, default = Point.Wall)
    visited_flag = [[False for _ in range(size[1])] for _ in range(size[0])] # 防止还在队列里的被重复打通，则会存在环路

    queue = [[start, (-1, -1)] for start in starts] # 当前点，前一个点
    while queue:
        cur, pre = queue.pop(random.randint(0, len(queue) - 1))
        if _valid_neighbour(m, cur) < 1:
            continue
        m.data[cur[0]][cur[1]] = Point.Chan
        visited_flag[cur[0]][cur[1]] = True

        if pre[0] != -1 and cur[0] == pre[0]:
            m.data[cur[0]][min(cur[1], pre[1]) + 1] = Point.Chan
        elif pre[1] != -1 and cur[1] == pre[1]:
            m.data[min(cur[0], pre[0]) + 1][cur[1]] = Point.Chan

        if callback is not None:
            callback(m.data)

        udlr = [(cur[0] - 2, cur[1]), (cur[0] + 2, cur[1]), (cur[0], cur[1] - 2), (cur[0], cur[1] + 2),]
        random.shuffle(udlr)
        for p in udlr:
            if 0 <= p[0] < size[0] and 0 <= p[1] < size[1] and m.data[p[0]][p[1]] == Point.Wall and not visited_flag[p[0]][p[1]]:
                queue.append([p, cur])
                visited_flag[p[0]][p[1]] = True # 目标点已经追加过

    # start 和 end 之间差非偶数的情况下无法联通，需要链接终点到最近的可行点
    m = _connect_points(m, ends)

    for p in starts:
        m.data[p[0]][p[1]] = Point.Start
    for p in ends:
        m.data[p[0]][p[1]] = Point.End
    return m


def kruskalg(size, starts, ends, callback = None):
    '''随机Kruskal算法 (并查集)
    1.创建所有墙的列表（除了四边），并且创建所有单元的集合，每个集合中只包含一个单元。
    2.随机从墙的列表中选取一个，取该墙两边分隔的两个单元
        1.两个单元属于不同的集合，则将去除当前的墙，把分隔的两个单元连同当前墙三个点作为一个单元集合；并将当前选中的墙移出列表
        2.如果属于同一个集合，则直接将当前选中的墙移出列表
    3.不断重复第 2 步，直到所有墙都检测过
    该算法同样不会出现明显的主路，岔路也比较多
    Pyhthon 实现可以参考(代码太难看): https://blog.csdn.net/cj12345657582255/article/details/115605195
    '''
    m = Map(size, default = Point.Wall)
    print("Unrelized method kruskalg")
    traceback.print_stack()
    for p in starts:
        m.data[p[0]][p[1]] = Point.Start
    for p in ends:
        m.data[p[0]][p[1]] = Point.End
    return m


if __name__ == '__main__':
    import doctest
    doctest.testmod()  # verbose=True shows the output
    
    # generate = primg # dfsg, primg, kruskalg
    # callback = lambda m: print('\n'.join(["-----", ] + [' '.join([p.name[0] for p in row]) for row in m]))
    # m = generate((3, 3), [(0, 0),], [(2, 2),], callback=callback)
    # print(m)
    # m = generate((6, 6), [(0, 0),], [(5, 5),], callback=callback)
    # print(m)
    # m = generate((60, 60), [(0, 0),], [(45, 51),])
    # print(m)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2023-01-17 22:38:10
# @Author  : HeLiang (helianghit@foxmail.com)
# @Link    : https://github.com/HeLiangHIT
# 此处r主要是为了支持在 doctest 命令中使用 \n
r'''
路径搜索，提供bfs(最短路)和dfs两种方法，入参相同，使用示例如下:
>>> m = Map((3, 3)); m.data[1][1] = Point.Wall; m.data[0][0] = Point.Start; m.data[1][2] = Point.End
>>> paths = bfs(m.data, [(0, 0),], [(1, 2),], callback=lambda m: print('\n'.join(["-----", ] + [' '.join([p.name[0] for p in row]) for row in m]))) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
-----
V C C
C W E
C C C
-----
V N C
V W E
C C C
...
V W N
V V C
-----
V V V
V W V
V V N
>>> print(paths) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
[[(0, 0), (0, 1), (0, 2), (1, 2)]]

>>> paths = dfs(m.data, [(0, 0),], [(1, 2),], callback=lambda m: print('\n'.join(["-----", ] + [' '.join([p.name[0] for p in row]) for row in m]))) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
-----
V C C
C W E
C C C
-----
V C C
V W E
...
V V V
-----
V C C
V W V
V V V
>>> print(paths) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
[[(0, 0), (1, 0), (2, 0), (2, 1), (2, 2), (1, 2)]]
>>> paths = a_star(m.data, [(0, 0),], [(1, 2),], callback=lambda m: print('\n'.join(["-----", ] + [' '.join([p.name[0] for p in row]) for row in m]))) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
-----
V C C
C W E
C C C
-----
V V C
...
V V V
N W V
C C C
>>> print(paths) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
[[(0, 0), (0, 1), (0, 2), (1, 2)]]
'''
import sys
import os
import copy
import traceback
from queue import PriorityQueue
from .map import Point, Map


def _valid_check(mdata, srcs, targets):
    '''输入有效性检查'''
    if not (isinstance(mdata, list) and isinstance(mdata[0], list)):
        print(f"input parameter mdata is illegal(must be 2-dimension): {mdata}")
        traceback.print_stack()
        return False
    if not all([isinstance(p, Point) for row in mdata for p in row]): # 所有成员都是 Point 子类
        print(f"input parameter mdata is illegal(must be all Point type): {mdata}")
        traceback.print_stack()
        return False
    if not (isinstance(srcs, list) and isinstance(srcs[0], tuple) and isinstance(srcs[0][0], int)):
        print(f"input parameter srcs is illegal(must be list of coordinate): {srcs}")
        traceback.print_stack()
        return False
    if not (isinstance(targets, list) and isinstance(targets[0], tuple) and isinstance(targets[0][0], int)):
        print(f"input parameter targets is illegal(must be list of coordinate): {targets}")
        traceback.print_stack()
        return False
    # TODO: srcs/targets 是否超过边界
    return True


def _backtrace(pre_points, srcs, cur):
    '''回溯从srcs到cur的路径'''
    path = [cur, ]
    if cur not in pre_points: # 当前点(目标点)和起始点重复的特殊情况
        path.append(cur)
        return path[::-1]
    while pre_points[cur] not in srcs: # 遍历找到从源到当前点的路径
        cur = pre_points[cur]
        path.append(cur)
    path.append(pre_points[cur])
    return path[::-1]


def bfs(mdata, srcs, targets, max_steps = 20000, max_paths = 1, callback = None):
    r'''
    多点对多点的BFS最短路算法
    NOTE: mdata 为二维数组，入参 srcs 和 targets 需为数组实际下标； callback 为每次迭代计算的回调，用于实时在视图中刷新访问情况
    '''
    if not _valid_check(mdata, srcs, targets):
        return []
    width = len(mdata[0]) # 注意数组的保存是从上到下、从左到右
    height = len(mdata)
    _mdata = copy.deepcopy(mdata) # 避免直接修改实参
    
    q = srcs[:] # 待探索的点，避免直接修改实参
    avaliable_paths = [] # 可达结果路径列表
    targets_set = set(targets) # 避免目标重复和直接修改实参

    steps = 0 # 计算次数，避免超时
    pre_points = {} # 前序索引，用于回溯

    while q and targets_set:
        cur, *q = q
        steps += 1
        _mdata[cur[0]][cur[1]] = Point.Visited
        if callback is not None:
            callback(_mdata)
        if cur in targets_set: # 找到目标
            avaliable_paths.append(_backtrace(pre_points, srcs, cur)) # 保存最短路径
            # targets_set.remove(cur) # 移除目标，此处无需移除，允许下一条路径到达该目标
        else: # 非目标节点，增加相邻的合法节点
            udlr = [(cur[0] - 1, cur[1]), (cur[0] + 1, cur[1]), (cur[0], cur[1] - 1), (cur[0], cur[1] + 1),]
            qadd = [p for p in udlr if 0 <= p[0] < height and 0 <= p[1] < width and _mdata[p[0]][p[1]] in (Point.Chan, Point.End)] # NOTE: 终点也可选
            for p in qadd: # 记录目标点是从当前位置过去的
                _mdata[p[0]][p[1]] = Point.NxtVisit # 已经走过，设置即将访问的点避免再次被加入访问队列
                if p not in pre_points: # 增加标记前进路线，避免重复增加
                    pre_points[p] = cur
            q.extend(qadd)
        if len(avaliable_paths) >= max_paths:
            break
        if steps >= max_steps:
            break
    return avaliable_paths


def _dfs(mdata, cur, targets, iter_status, max_steps, max_paths, callback):
    width = len(mdata[0]) # 注意数组的保存是从上到下、从左到右
    height = len(mdata)

    pre = mdata[cur[0]][cur[1]]
    mdata[cur[0]][cur[1]] = Point.Visited
    iter_status["steps"] += 1
    iter_status["cur_path"].append(cur)

    # 边界条件
    if len(iter_status["avaliable_paths"]) >= max_paths or iter_status["steps"] >= max_steps:
        return False
    if callback is not None:
        callback(mdata)
    if cur in targets:
        iter_status["avaliable_paths"].append(iter_status["cur_path"][:])
        return True

    udlr = [(cur[0] - 1, cur[1]), (cur[0] + 1, cur[1]), (cur[0], cur[1] - 1), (cur[0], cur[1] + 1),]
    qadd = [p for p in udlr if 0 <= p[0] < height and 0 <= p[1] < width and mdata[p[0]][p[1]] in (Point.Chan, Point.End)]
    for p in qadd:
        _dfs(mdata, p, targets, iter_status, max_steps, max_paths, callback)

    iter_status["cur_path"].remove(cur)
    mdata[cur[0]][cur[1]] = pre


def dfs(mdata, srcs, targets, max_steps = 50000, max_paths = 1, callback = None):
    '''
    多点对多点的DFS最短路算法
    NOTE: mdata 为二维数组，入参 srcs 和 targets 需为数组实际下标； callback 为每次迭代计算的回调，用于实时在视图中刷新访问情况
    '''
    pre_stack_limit = sys.getrecursionlimit()
    sys.setrecursionlimit(10000)
    if not _valid_check(mdata, srcs, targets):
        return []
    _mdata = copy.deepcopy(mdata) # 避免直接修改实参
    iter_status = {"steps": 0, # 访问步数
                   "cur_path": [], # 当前访问路径
                   "avaliable_paths": [], # 可达结果路径列表
                  }

    for p in srcs:
        _dfs(_mdata, p, targets, iter_status, max_steps, max_paths, callback)

    sys.setrecursionlimit(pre_stack_limit)
    return iter_status["avaliable_paths"]


def _calc_cost(cur, targets):
    # 计算当前点到目标点的最短距离
    dist = [abs(cur[0] - p[0]) + abs(cur[1] - p[1]) for p in targets]
    return min(dist)

def a_star(mdata, srcs, targets, max_steps = 50000, max_paths = 1, callback = None):
    '''
    多点对多点的A*最短路算法 -> BFS 算法改进为以距离优先出
    NOTE: mdata 为二维数组，入参 srcs 和 targets 需为数组实际下标； callback 为每次迭代计算的回调，用于实时在视图中刷新访问情况
    '''
    if not _valid_check(mdata, srcs, targets):
        return []
    width = len(mdata[0]) # 注意数组的保存是从上到下、从左到右
    height = len(mdata)
    _mdata = copy.deepcopy(mdata) # 避免直接修改实参
    
    q = PriorityQueue() # put by (priority_number, data) and lowest priority retrieved first
    for p in srcs:
        q.put((0, p))
    avaliable_paths = [] # 可达结果路径列表
    targets_set = set(targets) # 避免目标重复和直接修改实参

    steps = 0 # 计算次数，避免超时
    pre_points = {} # 前序索引，用于回溯

    while not q.empty() and targets_set: # 注意 q == True
        _, cur = q.get()
        steps += 1
        _mdata[cur[0]][cur[1]] = Point.Visited
        if callback is not None:
            callback(_mdata)
        if cur in targets_set: # 找到目标
            avaliable_paths.append(_backtrace(pre_points, srcs, cur)) # 保存最短路径
            # targets_set.remove(cur) # 移除目标，此处无需移除，允许下一条路径到达该目标
        else: # 非目标节点，增加相邻的合法节点
            udlr = [(cur[0] - 1, cur[1]), (cur[0] + 1, cur[1]), (cur[0], cur[1] - 1), (cur[0], cur[1] + 1),]
            qadd = [p for p in udlr if 0 <= p[0] < height and 0 <= p[1] < width and _mdata[p[0]][p[1]] in (Point.Chan, Point.End)] # NOTE: 终点也可选
            # NOTE: A* 算法和BFS算法的主要差异点就在这里
            for p in qadd: # 记录目标点是从当前位置过去的
                q.put((_calc_cost(p, targets), p))
                _mdata[p[0]][p[1]] = Point.NxtVisit # 已经走过，设置即将访问的点避免再次被加入访问队列
                if p not in pre_points: # 增加标记前进路线，避免重复增加
                    pre_points[p] = cur
        if len(avaliable_paths) >= max_paths:
            break
        if steps >= max_steps:
            break
    return avaliable_paths



if __name__ == '__main__':
    import doctest
    doctest.testmod()  # verbose=True shows the output

    # m = Map((15, 20), default = Point.Wall)
    # m.data[0][0] = Point.Start
    # m.data[0][1] = Point.Chan
    # m.data[14][19] = Point.End
    # answers = a_star(m.data, [(0, 0),], [(14, 19),])
    # print(answers)

    # data = [[Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall], 
    #         [Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall], 
    #         [Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Chan, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Chan, Point.Chan, Point.Wall], 
    #         [Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Chan, Point.Wall], 
    #         [Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Wall, Point.Chan, Point.Wall], 
    #         [Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall], 
    #         [Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Chan, Point.Chan, Point.Chan, Point.Wall, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Wall], 
    #         [Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall], 
    #         [Point.Wall, Point.Wall, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Wall, Point.Chan, Point.Chan, Point.Chan, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall], 
    #         [Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall], 
    #         [Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall], 
    #         [Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall], 
    #         [Point.Chan, Point.Chan, Point.Chan, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Wall, Point.Wall, Point.Wall], 
    #         [Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Wall, Point.Wall, Point.Wall], 
    #         [Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Chan, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Wall, Point.Chan, Point.Chan, Point.Chan, Point.Chan]
    #     ]
    # start = (0, 0)
    # end = (14, 19)
    # ans = bfs(data, [start, ], [end, ])
    # print(ans)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2023-01-31 16:02:56
# @Author  : HeLiang (helianghit@foxmail.com)
# @Link    : https://github.com/HeLiangHIT
# @Ref     : https://zhuanlan.zhihu.com/p/276461821?utm_source=wechat_session
#            https://www.zhihu.com/question/20609631
import os
from setuptools import setup


def sub_path(rpath):
    script_dir = os.path.abspath(os.path.dirname(__file__))
    return os.path.sep.join([script_dir, rpath])


def read_file(path):
    with open(path, 'r', encoding='utf-8') as f:
        return f.read()
    return ''


def get_latest_tag():
    from git.repo import Repo # 需在发布机上 pip install gitpython
    tags = Repo().tags
    if len(tags) > 0:
        return tags[-1].name
    else:
        return 'v0.0.0' # 初始版本号


def list_packages():
    import pkgutil
    return [name for _, name, ispkg in pkgutil.walk_packages([sub_path('.'),] ) if ispkg]


def list_modules():
    import pkgutil
    return [name for _, name, ispkg in pkgutil.walk_packages([sub_path('.'),] ) if not ispkg]


def find_files(directory, strip = os.path.abspath(os.path.dirname(__file__))):
  result = []
  for root, dirs, files in os.walk(directory):
    for filename in files:
      filename = os.path.join(root, filename)
      result.append(os.path.relpath(filename, strip))
  return result


setup(
    name = 'magic_maze',
    version = get_latest_tag(),
    url = 'https://github.com/HeLiangHIT/magic_maze',
    license = 'Mulan PSL v2',
    author = 'He Liang',
    author_email = 'heianghit@foxmail.com',
    keywords = 'maze pyqt',

    description = 'maze game developed by python',
    long_description_content_type = "text/markdown",
    long_description = read_file(sub_path('README.md')),

    packages = list_packages(),
    py_modules = list_modules(),
    install_requires = read_file(sub_path('requirement.txt')).split('\n'),
    entry_points = {'console_scripts': [
        'maze = main:main', # cmd entry points
    ]},
    include_package_data = True, 
    data_files = [('', ['README.md', 'LICENSE']), # try package_data instead
                  ('script', find_files(sub_path("script"))),
                  ('demo', find_files(sub_path("demo"))),
                  ('doc', find_files(sub_path("doc"))),
                  ('ui/resources', find_files(sub_path("ui/resources"))),],

    zip_safe = False, # avoid error when uninstall on windows

    classifiers = [
        'Programming Language :: Python :: 3',
        'Development Status :: 5 - Production/Stable',
        'Natural Language :: Chinese (Simplified)',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop', 
        'Operating System :: OS Independent',
        'Topic :: Games/Entertainment',
        'License :: OSI Approved :: Mulan Permissive Software License v2 (MulanPSL-2.0)',
    ],
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Date    : 2023-01-18 16:30:36
# @Author  : HeLiang (helianghit@foxmail.com)
# @Link    : https://github.com/HeLiangHIT
# @ref     : https://doc.qt.io/qtforpython/PySide6/QtWidgets/QGraphicsView.html#qgraphicsview

'''
pyqt 编写的贪吃蛇界面
'''
import os
import sys
import time
from functools import reduce
from math import ceil
import copy
from PyQt5 import QtCore, QtGui, QtWidgets, Qt, QtTest
from PyQt5.QtWidgets import QApplication
from algorithm.map import Point, Map
from algorithm.search import bfs, dfs, a_star
from algorithm.generate import dfsg, primg, kruskalg


_script_dir = os.path.dirname(os.path.realpath(__file__))


class MazeMain(QtWidgets.QWidget):
    DEFAULT = (15, 20) # 默认迷宫尺寸
    MAX = (300, 400) # 最大支持的迷宫尺寸
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.init_ui()
        self.gen_btn.clicked.connect(self.generate_maze)
        self.draw_btn.clicked.connect(self.draw_maze)
        self.run_btn.clicked.connect(self.run_maze)
        self.auto_btn.clicked.connect(self.auto_maze)
        self.load_btn.clicked.connect(self.load_maze)
        self.save_btn.clicked.connect(self.save_maze)
        # 自动修改终点位置
        self.maze_height.textChanged.connect(lambda x: self.end_x.setText("%s" % (int(x) - 1)))
        self.maze_width.textChanged.connect(lambda x: self.end_y.setText("%s" % (int(x) - 1)))

    def load_maze(self):
        path, fileType = QtWidgets.QFileDialog.getOpenFileName(self, "选择文件", './demo', "Map Files(*.map)")
        if not path:
            return
        with open(path, 'r') as f:
            self.maze_map, ok = Map().from_json(f.read())
            if not ok:
                QtWidgets.QMessageBox.information(self, "文件无效", "无法解析该文件", QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.Yes)
                return
        self.size, self.start, self.end = self.maze_map.size, self.maze_map.start[0], self.maze_map.end[0]
        [e.setText(str(v)) for e, v in zip((self.maze_height, self.maze_width, self.start_x, self.start_y, self.end_x, self.end_y), self.size + self.start + self.end)]
        self.maze.new_maze(self.maze_map)

    def save_maze(self):
        path, _ = QtWidgets.QFileDialog.getSaveFileName(self, 'Save File', f'./demo/{self.size[0]}X{self.size[1]}.map')
        if not path:
            return
        with open(path,'w') as f:
            f.write(self.maze_map.to_json())

    def draw_maze(self):
        # 鼠标左键绘制通道/墙体
        if not self.is_drawing:
            self._start_draw()
        else:
            self._finish_draw()
        self.is_drawing = not self.is_drawing
    def _finish_draw(self):
        if self.draw_btn.text().count("完成") > 0:
            self.maze_map = self.maze.get_map()
        self.maze.new_maze(self.maze_map)
        [btn.setDisabled(False) for btn in (self.gen_btn, self.run_btn, self.auto_btn)]
        self.draw_btn.setText("绘制迷宫")
    def _start_draw(self):
        [btn.setDisabled(True) for btn in (self.gen_btn, self.run_btn, self.auto_btn)]
        self.draw_btn.setText("无效迷宫，取消绘制")
        def fresh_and_check(m):
            answers = a_star(m.data, [self.start,], [self.end,])
            self.draw_btn.setText(("无效迷宫，取消绘制", "有效迷宫，完成绘制")[len(answers) > 0])
        self.size, self.start, self.end = self._parser_input_text()
        m = Map(self.size, default = Point.Wall)
        m.data[self.start[0]][self.start[1]] = Point.Start
        m.data[self.end[0]][self.end[1]] = Point.End
        self.maze.new_maze(m, callback = fresh_and_check, mode = "draw") # 注册绘制的回调函数

    def run_maze(self):
        # 鼠标左键绘制路线，右键消除路线 - 走迷宫的时候重新生成或者自动走都不会影响绘制，可以不禁用其它按钮
        def fresh_and_check(m):
            # 检查终点附近是不是已经走到了
            udlr = [(self.end[0] - 1, self.end[1]), (self.end[0] + 1, self.end[1]), (self.end[0], self.end[1] - 1), (self.end[0], self.end[1] + 1),]
            visited_nb = reduce(lambda x, y: x + y, 
                                [0 <= q[0] < self.size[0] and 0 <= q[1] < self.size[1] and m.data[q[0]][q[1]] in (Point.Start, Point.Visited) for q in udlr])
            if visited_nb >= 1:
                ndata = [[Point.Chan if p in (Point.Visited, Point.Start, Point.End) else Point.Wall for p in row] for row in m.data] # 用于根据走的路径寻找答案
                answers = bfs(ndata, [self.start, ], [self.end, ])
                self.maze.new_maze(self.maze_map, answers[0]) # 重绘答案
                QtWidgets.QMessageBox.information(self, "恭喜通关", "恭喜通关", QtWidgets.QMessageBox.Yes, QtWidgets.QMessageBox.Yes)
        self.maze.new_maze(self.maze_map, callback = fresh_and_check, mode = "run") # 注册绘制的回调函数

    def generate_maze(self):
        idx = self.gen_combo.currentIndex()
        generate = {0: primg, 1: dfsg}[idx]
        self.size, self.start, self.end = self._parser_input_text()
        self.maze_map = generate(self.size, [self.start, ], [self.end, ], callback = self._fresh_callback if self.visible_check.isChecked() else None)
        self.maze.new_maze(self.maze_map)

    def auto_maze(self):
        idx = self.auto_combo.currentIndex()
        search = {0: bfs, 1: dfs, 2: a_star}[idx]
        answers = search(self.maze_map.data, [self.start,], [self.end,], callback = self._fresh_callback if self.visible_check.isChecked() else None)
        answer = answers[0] if len(answers) > 0 else []
        self.maze.new_maze(self.maze_map, answer)

    def _fresh_callback(self, data, interval = 0.01):
        speed = self.speed_slider.value()
        interval = (1e-7, 1e-4, 1e-1)[(speed < 33) + (speed < 66)] # 三个档慢、中、快
        m = Map().from_data(data)
        self.maze.update_maze(m)
        # QApplication.processEvents() # 强制刷新界面 - 性能较差，改为在组件中 repaint
        # time.sleep(interval) # 会卡UI
        QtTest.QTest.qWait(interval * 1000) # 单位 msecs, 不卡UI

    def _parser_input_text(self):
        size = (int(self.maze_height.text()), int(self.maze_width.text()))
        start = (int(self.start_x.text()), int(self.start_y.text()))
        end = (int(self.end_x.text()), int(self.end_y.text()))
        return size, start, end

    def init_ui(self):
        self.resize(1024, 624)
        self.setWindowTitle("Maze")
        self.setWindowIcon(QtGui.QIcon(os.path.sep.join([_script_dir, "resources", "maze.png"])))
        h_layout_main = QtWidgets.QHBoxLayout(self)
        h_layout_main.setContentsMargins(2, 2, 2, 2)

        self.maze = MazeWidget()
        self.maze.setStyleSheet("margin:0px; padding:0px; background:rgba(0,0,0,0); border:none;")
        h_layout_main.addWidget(self.maze)

        v_layout_right = QtWidgets.QVBoxLayout()
        title = QtWidgets.QLabel(self)
        title.setAlignment(QtCore.Qt.AlignCenter)
        title.setStyleSheet("QLabel{font-size:30pt;font-style:bold;text-align:center;}")
        title.setText("Maze")
        v_layout_right.addWidget(title)

        gb = QtWidgets.QGroupBox("通用输入", self)
        v_layout_input = QtWidgets.QVBoxLayout()
        h_layout_size = QtWidgets.QHBoxLayout()
        size_text = QtWidgets.QLabel(gb)
        size_text.setText("尺寸: ")
        h_layout_size.addWidget(size_text)
        self.maze_height = QtWidgets.QLineEdit(gb)
        self.maze_height.setText("%s" % self.DEFAULT[0])
        self.maze_height.setValidator(QtGui.QIntValidator(3, self.MAX[0]))
        h_layout_size.addWidget(self.maze_height)
        size_spliter = QtWidgets.QLabel(gb)
        size_spliter.setText("X")
        h_layout_size.addWidget(size_spliter)
        self.maze_width = QtWidgets.QLineEdit(gb)
        self.maze_width.setText("%s" % self.DEFAULT[1])
        self.maze_width.setValidator(QtGui.QIntValidator(3, self.MAX[1]))
        h_layout_size.addWidget(self.maze_width)
        v_layout_input.addLayout(h_layout_size)
        
        h_layout_start = QtWidgets.QHBoxLayout()
        start_text = QtWidgets.QLabel(gb)
        start_text.setText("起点: ")
        h_layout_start.addWidget(start_text)
        self.start_x = QtWidgets.QLineEdit(gb)
        self.start_x.setText("0")
        self.start_x.setValidator(QtGui.QIntValidator(0, self.MAX[0] - 1))
        h_layout_start.addWidget(self.start_x)
        start_spliter = QtWidgets.QLabel(gb)
        start_spliter.setText(",")
        h_layout_start.addWidget(start_spliter)
        self.start_y = QtWidgets.QLineEdit(gb)
        self.start_y.setText("0")
        self.start_y.setValidator(QtGui.QIntValidator(0, self.MAX[1] - 1))
        h_layout_start.addWidget(self.start_y)
        v_layout_input.addLayout(h_layout_start)

        h_layout_end = QtWidgets.QHBoxLayout()
        end_text = QtWidgets.QLabel(gb)
        end_text.setText("终点: ")
        h_layout_end.addWidget(end_text)
        self.end_x = QtWidgets.QLineEdit(gb)
        self.end_x.setText("%s" % (self.DEFAULT[0] - 1))
        self.end_x.setValidator(QtGui.QIntValidator(0, self.MAX[0] - 1))
        h_layout_end.addWidget(self.end_x)
        end_spliter = QtWidgets.QLabel(gb)
        end_spliter.setText(",")
        h_layout_end.addWidget(end_spliter)
        self.end_y = QtWidgets.QLineEdit(gb)
        self.end_y.setText("%s" % (self.DEFAULT[1] - 1))
        self.end_y.setValidator(QtGui.QIntValidator(0, self.MAX[1] - 1))
        h_layout_end.addWidget(self.end_y)
        v_layout_input.addLayout(h_layout_end)

        gb.setLayout(v_layout_input)
        v_layout_right.addWidget(gb)

        gb = QtWidgets.QGroupBox("自动操作", self)
        v_layout_auto = QtWidgets.QVBoxLayout()

        h_layout_visible = QtWidgets.QHBoxLayout()
        self.visible_check = QtWidgets.QCheckBox("显示/速度:", self)
        self.visible_check.setChecked(False) # isChecked
        h_layout_visible.addWidget(self.visible_check)
        self.speed_slider = QtWidgets.QSlider(QtCore.Qt.Horizontal, self)
        self.speed_slider.setTickInterval(33)
        self.speed_slider.setValue(50)
        self.speed_slider.setTickPosition(QtWidgets.QSlider.TicksBothSides)
        h_layout_visible.addWidget(self.speed_slider) # value()
        v_layout_auto.addLayout(h_layout_visible)

        h_layout_gen = QtWidgets.QHBoxLayout()
        self.gen_combo = QtWidgets.QComboBox(self)
        self.gen_combo.addItem("Prim算法")
        self.gen_combo.addItem("深度优先")
        # self.gen_combo.addItem("Kruskal算法") # TODO: 待开发
        h_layout_gen.addWidget(self.gen_combo)
        self.gen_btn = QtWidgets.QPushButton(self)
        self.gen_btn.setText("生成迷宫")
        h_layout_gen.addWidget(self.gen_btn)
        v_layout_auto.addLayout(h_layout_gen)

        h_layout_auto = QtWidgets.QHBoxLayout()
        self.auto_combo = QtWidgets.QComboBox(self)
        self.auto_combo.addItem("BFS")
        self.auto_combo.addItem("DFS")
        self.auto_combo.addItem("A*")
        h_layout_auto.addWidget(self.auto_combo)
        self.auto_btn = QtWidgets.QPushButton(self)
        self.auto_btn.setText("走迷宫")
        h_layout_auto.addWidget(self.auto_btn)
        v_layout_auto.addLayout(h_layout_auto)

        gb.setLayout(v_layout_auto)
        v_layout_right.addWidget(gb)

        gb = QtWidgets.QGroupBox("手动操作", self)
        v_layout_handle = QtWidgets.QVBoxLayout()
        self.draw_btn = QtWidgets.QPushButton(self)
        self.draw_btn.setText("绘制迷宫")
        self.is_drawing = False
        v_layout_handle.addWidget(self.draw_btn)
        self.run_btn = QtWidgets.QPushButton(self)
        self.run_btn.setText("走迷宫")
        v_layout_handle.addWidget(self.run_btn)
        notice_lbl = QtWidgets.QLabel(self)
        notice_lbl.setStyleSheet("QLabel{font-size:6pt;font-style:italic;text-align:center;}")
        notice_lbl.setText("左键绘制/右键清除/触摸翻转")
        notice_lbl.setAlignment(QtCore.Qt.AlignCenter)
        v_layout_handle.addWidget(notice_lbl)

        gb.setLayout(v_layout_handle)
        v_layout_right.addWidget(gb)

        gb = QtWidgets.QGroupBox("加载/保存", self)
        h_layout_handle = QtWidgets.QHBoxLayout()
        self.load_btn = QtWidgets.QPushButton(self)
        self.load_btn.setText("加载")
        h_layout_handle.addWidget(self.load_btn)
        self.save_btn = QtWidgets.QPushButton(self)
        self.save_btn.setText("保存")
        h_layout_handle.addWidget(self.save_btn)

        gb.setLayout(h_layout_handle)
        v_layout_right.addWidget(gb)

        right_widget = QtWidgets.QWidget() # 为了控制右侧宽度
        right_widget.setFixedWidth(224)
        right_widget.setLayout(v_layout_right)
        h_layout_main.addWidget(right_widget)

        # self.setWindowFlags(QtCore.Qt.WindowCloseButtonHint)
        # self.setFixedSize(self.width(), self.height())
        self.generate_maze()

class MazeWidget(QtWidgets.QWidget):
    ColorMap = {
        Point.Wall: QtCore.Qt.black,
        Point.Start: QtCore.Qt.blue,
        Point.End: QtCore.Qt.blue,
        Point.Visited: QtCore.Qt.cyan,
        Point.NxtVisit: QtCore.Qt.gray,
        }
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setAttribute(QtCore.Qt.WA_AcceptTouchEvents, True) # 允许触摸响应
        self.m = None

    def new_maze(self, maze_map, answer = None, callback = None, mode = "display"):
        # 新的迷宫数据
        self.m = copy.deepcopy(maze_map)
        self.answer = answer if answer is not None else []
        self.col_num = maze_map.size[1] # 列 -> 二维
        self.row_num = maze_map.size[0]
        self.callback = callback
        self.diff = None
        self.pix = None
        self.mode = mode
        self.repaint() # 强制刷新迷宫

    def update_maze(self, maze_map):
        # 增量刷新迷宫显示区域
        if self.m is None or self.m.size != maze_map.size:
            self.new_maze(maze_map) # 首次设置
        else:
            self.diff = self.m.diff(maze_map)
            self.m = copy.deepcopy(maze_map)
            self.pix = self.grab(QtCore.QRect(0, 0, self.width(), self.height())) # 避免仅显示绘制部分，作为二级缓存显示
            self.repaint()

    def get_map(self):
        return copy.deepcopy(self.m)

    def paintEvent(self, e):
        super().paintEvent(e)
        painter = QtGui.QPainter(self)
        self.col_len = self.width() / self.col_num
        self.row_len = self.height() / self.row_num
        if self.diff and self.pix:
            painter.drawPixmap(0, 0, self.width(), self.height(), self.pix)
            self._draw_diff(painter)
        else:
            self._draw_all(painter)
            self._draw_answer(painter)

    def _draw_all(self, painter):
        # 绘制迷宫
        for m in range(self.col_num):
            for n in range(self.row_num):
                color = self.ColorMap[self.m.data[n][m]] if self.m.data[n][m] in self.ColorMap.keys() else QtCore.Qt.white
                painter.setPen(color)
                painter.setBrush(color)
                painter.drawRect(ceil(m * self.col_len), ceil(n * self.row_len), ceil(self.col_len), ceil(self.row_len))
                if self.m.data[n][m] == Point.Start:
                    painter.drawImage(QtCore.QRect(ceil(m * self.col_len), ceil(n * self.row_len), ceil(self.col_len), ceil(self.row_len)), 
                        QtGui.QImage(os.path.sep.join([_script_dir, "resources", "run.png"])))
                if self.m.data[n][m] == Point.End:
                    painter.drawImage(QtCore.QRect(ceil(m * self.col_len), ceil(n * self.row_len), ceil(self.col_len), ceil(self.row_len)), 
                        QtGui.QImage(os.path.sep.join([_script_dir, "resources", "flag.png"])))

    def _draw_diff(self, painter):
        # 绘制增量
        for n, m, v in self.diff:
            color = self.ColorMap[v] if v in self.ColorMap.keys() else QtCore.Qt.white
            painter.setPen(color)
            painter.setBrush(color)
            painter.drawRect(ceil(m * self.col_len), ceil(n * self.row_len), ceil(self.col_len), ceil(self.row_len))

    def _draw_answer(self, painter):
        # 绘制答案
        for n, m in self.answer[1:-1]: # 起止点有专门的标记，不用重复绘制
            painter.setPen(QtCore.Qt.yellow)
            painter.setBrush(QtCore.Qt.red)
            painter.drawEllipse(ceil(m * self.col_len + self.col_len / 4), ceil(n * self.row_len + self.row_len / 4), 
                                ceil(self.col_len / 2), ceil(self.row_len / 2))

    def event(self, e):
        if e.type() not in (QtGui.QTouchEvent.TouchBegin, QtGui.QTouchEvent.TouchUpdate, QtGui.QTouchEvent.TouchEnd):
            return super().event(e)
        if self.mode == "run":
            self._flip_cell_for_run_maze(e.touchPoints()[0].pos(), QtCore.Qt.LeftButton)
        if self.mode == "draw":
            self._flip_cell_for_draw_maze(e.touchPoints()[0].pos(), QtCore.Qt.LeftButton)
        return True

    def mousePressEvent(self, e):
        self.mouseMoveEvent(e)

    def mouseMoveEvent(self, e):
        if self.mode == "run":
            self._flip_cell_for_run_maze(e.pos(), e.buttons())
        if self.mode == "draw":
            self._flip_cell_for_draw_maze(e.pos(), e.buttons())

    def _flip_cell_for_draw_maze(self, pos, btn):
        # 获取坐标所在迷宫位置并标记已访问，注意此处QT的坐标和矩阵是反的
        # 鼠标移动时翻转墙体和通道 - 支持触屏翻转
        row = int(pos.y() / self.row_len)
        col = int(pos.x() / self.col_len)
        # 确保鼠标没有移出边界
        if not (0 <= row < self.row_num and 0 <= col < self.col_num):
            return
        if self.m.data[row][col] == Point.Wall and btn == QtCore.Qt.LeftButton:
            self.m.data[row][col] = Point.Chan
        elif self.m.data[row][col] == Point.Chan and btn == QtCore.Qt.RightButton:
            self.m.data[row][col] = Point.Wall
        self.repaint() # 强制刷新迷宫
        if self.callback:
            self.callback(self.m)
        
    def _flip_cell_for_run_maze(self, pos, btn):
        # 获取坐标所在迷宫位置并标记已访问，注意此处QT的坐标和矩阵是反的
        # 鼠标移动时左键走路线、右键清除路线 - 触屏走迷宫时可以不需要清除路线
        row = int(pos.y() / self.row_len)
        col = int(pos.x() / self.col_len)
        # 确保鼠标没有移出边界
        if not (0 <= row < self.row_num and 0 <= col < self.col_num):
            return
        # 确保附近有访问过或者在起点边上才能走通 - 无法支持倒着走
        udlr = [(row - 1, col), (row + 1, col), (row, col - 1), (row, col + 1),]
        visited_nb = reduce(lambda x, y: x + y, 
                              [0 <= q[0] < self.row_num and 0 <= q[1] < self.col_num and self.m.data[q[0]][q[1]] in (Point.Start, Point.Visited) for q in udlr])
        if visited_nb != 1:
            return
        if self.m.data[row][col] == Point.Chan and btn == QtCore.Qt.LeftButton:
            self.m.data[row][col] = Point.Visited
        elif self.m.data[row][col] == Point.Visited and btn == QtCore.Qt.RightButton:
            self.m.data[row][col] = Point.Chan
        self.repaint() # 强制刷新迷宫
        if self.callback:
            self.callback(self.m)


def main():
    app = QApplication(sys.argv)
    gui = MazeMain()
    gui.show()
    return app.exec_()


if __name__ == '__main__':
    main()

